import sys, os

import requests

import copy

import logging

from . import modelBase
from ...utils.input import getLabelIndex

import pyGrained.models.AlphaCarbon as proteinModel

class ENM(modelBase):
    """
    Component name: ENM
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 24/03/2023

    Elastic Network Model (ENM) model for proteins.

    """

    def __isValidPDB(self, id_pdb):
        """Checks if the PDB ID is valid."""
        return len(id_pdb) == 4 and id_pdb.isalnum()

    def __downloadPDB(self, id_pdb, file_path):
        url = f"https://files.rcsb.org/download/{id_pdb}.pdb"
        response = requests.get(url)

        if response.status_code == 200:
            with open(file_path, 'wb') as file:
                file.write(response.content)
            self.logger.info(f"PDB {id_pdb} downloaded successfully.")
        else:
            self.logger.error(f"Error downloading the PDB {id_pdb}. Please verify the PDB ID.")
            raise RuntimeError("Error downloading the PDB,")


    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"PDB",
                                                "centerInput",
                                                "SASA",
                                                "aggregateChains",
                                                "K",
                                                "enmCut"},
                         requiredParameters  = {"PDB"},
                         definedSelections   = {"particleId","forceField"},
                         **params)

        ############################################################
        ######################  Set up model  ######################
        ############################################################

        #Check if PDB is a file or a PDB ID.
        #PDB is considered a file if it has a .pdb or .pqr extension.
        #If it is a file, it is loaded as a PDB file.
        #If it is a PDB ID, PDB file is downloaded from the RCSB PDB database.

        downloadedPDB = False
        if params["PDB"].split(".")[-1] in ["pdb","pqr"]:
            inputPDBfilePath = params["PDB"]
        else:
            pdb_id = params["PDB"]
            if self.__isValidPDB(pdb_id):
                downloadedPDB = True
                inputPDBfilePath = f"tmp_{pdb_id}.pdb"
                self.__downloadPDB(params["PDB"], inputPDBfilePath)
            else:
                self.logger.error("The PDB ID is not valid")
                raise RuntimeError("Invalid PDB ID")


        enmParams = {"SASA":params.get("SASA",False),
                     "centerInput":params.get("centerInput",True),
                     "aggregateChains":params.get("aggregateChains",True),
                     "parameters": copy.deepcopy(params)}

        enm = proteinModel.ElasticNetworkModel(name = name,
                                               inputPDBfilePath = inputPDBfilePath,
                                               params = enmParams)
        if downloadedPDB:
            os.remove(inputPDBfilePath)
        ########################################################

        types = self.getTypes()
        modelTypes = enm.getTypes()

        for _,t in modelTypes.items():
            types.addType(**t)

        #Set model
        self.setState(enm.getState())
        self.setStructure(enm.getStructure())
        self.setForceField(enm.getForceField())


    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        if "forceField" in params:
            sel += self.getForceFieldSelection(params["forceField"])

        return sel




from VLMP.components.modelExtensions import modelExtensionBase

class helixBoundaries(modelExtensionBase):

    """
    Component name: helixBoundaries
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 03/06/2024

    :param helixPitch: pitch of the helix
    :type helixPitch: float
    :param helixRadius: radius of the helix
    :type helixRadius: float
    :param eps: helix sign, 1 for right-handed, -1 for left-handed
    :type eps: float
    :param nTurns: number of turns of the helix
    :type nTurns: int
    :param nPointsHelix: number of points to discretize the helix
    :type nPointsHelix: int
    :param helixInnerRadius: inner radius of the helix
    :type helixInnerRadius: float
    :param nx: number of points in the x direction
    :type nx: int
    :param ny: number of points in the y direction
    :type ny: int
    :param nz: number of points in the z direction
    :type nz: int
    :param K: spring constant for boundary conditions
    :type K: float

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"helixPitch","helixRadius",
                                                "eps","nTurns",
                                                "nPointsHelix","helixInnerRadius","K",
                                                "nx","ny","nz"},
                         requiredParameters  = {"helixPitch","helixRadius",
                                                "eps","nTurns",
                                                "nPointsHelix","helixInnerRadius","K",
                                                "nx","ny","nz"},
                         availableSelections = {"selection"},
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        helixPitch  = params["helixPitch"]
        helixRadius = params["helixRadius"]
        if not isinstance(helixPitch,float):
            self.logger.error("Invalid value for helixPitch. It must be a float")
            raise ValueError("Invalid value for helixPitch")
        if not isinstance(helixRadius,float):
            self.logger.error("Invalid value for helixRadius. It must be a float")
            raise ValueError("Invalid value for helixRadius")

        eps = params["eps"]
        eps = float(eps)

        if eps not in [-1.0,1.0]:
            self.logger.error("Invalid value for eps. It must be either 1.0 or -1.0")
            raise ValueError("Invalid value for eps")

        nTurns = params["nTurns"]
        #Check if nTurns is an integer
        if not isinstance(nTurns,int):
            self.logger.error("Invalid value for nTurns. It must be an integer")
            raise ValueError("Invalid value for nTurns")

        nPointsHelix = params["nPointsHelix"]
        #Check if nPointsHelix is an integer
        if not isinstance(nPointsHelix,int):
            self.logger.error("Invalid value for nPointsHelix. It must be an integer")
            raise ValueError("Invalid value for nPointsHelix")

        helixInnerRadius = params["helixInnerRadius"]
        #Check if helixInnerRadius is a float
        if not isinstance(helixInnerRadius,float):
            self.logger.error("Invalid value for helixInnerRadius. It must be a float")
            raise ValueError("Invalid value for helixInnerRadius")

        K = params["K"]
        #Check if K is a float
        if not isinstance(K,float):
            self.logger.error("Invalid value for K. It must be a float")
            raise ValueError("Invalid value for K")

        nx = params["nx"]
        ny = params["ny"]
        nz = params["nz"]

        #Check if nx, ny and nz are integers
        if not isinstance(nx,int):
            self.logger.error("Invalid value for nx. It must be an integer")
            raise ValueError("Invalid value for nx")
        if not isinstance(ny,int):
            self.logger.error("Invalid value for ny. It must be an integer")
            raise ValueError("Invalid value for ny")
        if not isinstance(nz,int):
            self.logger.error("Invalid value for nz. It must be an integer")
            raise ValueError("Invalid value for nz")

        if "selection" in params:
            self.setGroup("selection")

        ############################################################

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["External","HelixBoundaries"]
        extension[name]["parameters"] = {"helixPitch":helixPitch,
                                         "helixRadius":helixRadius,
                                         "eps":eps,
                                         "nTurns":nTurns,
                                         "nPointsHelix":nPointsHelix,
                                         "helixInnerRadius":helixInnerRadius,
                                         "nx":nx,"ny":ny,"nz":nz,
                                         "K":K}

        ############################################################

        self.setExtension(extension)



import sys, os

import copy

import logging

from . import modelBase
from ...utils.input import getLabelIndex

class MEMBRANE(modelBase):
    """
    Component name: MEMBRANE
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 22/05/2024

    Alpha carbon resolution membrane model.

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"position"},
                         requiredParameters  = set(),
                         definedSelections   = {"particleId"},
                         **params)

        ############################################################
        ######################  Set up model  ######################
        ############################################################

        # Now only POPC is implemented
        sigmaT = 7.111
        sigmaH = 0.65*sigmaT

        eps = 0.416
        w   = 9.867

        spacing  = sigmaT*0.1
        zSpacing = sigmaT*0.0

        lipidLen = 2.0*sigmaT + 3.0*spacing

        # Bond coefficients for POPC
        kH1_H2_POPC = 2.0*0.446
        kH2_T1_POPC = 2.0*1.073
        kT1_T2_POPC = 2.0*1.001
        kT2_T3_POPC = 2.0*0.443

        # Bond lengths for POPC
        b0_H1_H2_POPC = 5.580
        b0_H2_T1_POPC = 5.452
        b0_T1_T2_POPC = 5.050
        b0_T2_T3_POPC = 5.095

        # Angle coefficients for POPC
        kH1_H2_T1_POPC = 2.0*0.600
        kH2_T1_T2_POPC = 2.0*2.383
        kT1_T2_T3_POPC = 2.0*0.880

        # Angle equilibrium values for POPC
        theta0_H1_H2_T1_POPC = 3.142
        theta0_H2_T1_T2_POPC = 3.142
        theta0_T1_T2_T3_POPC = 3.142

        ########################################################

        types = self.getTypes()

        types.addType(name="H",radius=sigmaH/2.0,charge=0.0,mass=100.0)
        types.addType(name="T",radius=sigmaT/2.0,charge=0.0,mass=100.0)

        box = self.getEnsemble().getEnsembleComponent("box")

        # Create the membrane
        nLipidsX = box[0]/(sigmaT+spacing)
        nLipidsY = box[1]/(sigmaT+spacing)

        N = nLipidsX*nLipidsY

        # Create the lipids
        dx = sigmaT + spacing
        dy = sigmaT + spacing

        lipid = ["T","T","T","H","H"]

        lipidZ = [0.0]*len(lipid)
        lipidZ[0] = sigmaT/2.0 + zSpacing
        lipidZ[1] = lipidZ[0] + b0_H2_T1_POPC
        lipidZ[2] = lipidZ[1] + b0_T1_T2_POPC
        lipidZ[3] = lipidZ[2] + b0_T2_T3_POPC
        lipidZ[4] = lipidZ[3] + b0_T2_T3_POPC

        #Generate state
        state = {}
        state["labels"] = ["id","position"]
        state["data"]   = []

        #Generate structure
        structure = {}
        structure["labels"] = ["id","type","modelId"]
        structure["data"]   = []

        forceField = {}
        forceField["bond"] = {}
        forceField["bond"]["type"]       = ["Bond2", "Harmonic"]
        forceField["bond"]["labels"]     = ["id_i", "id_j", "K", "r0"]
        forceField["bond"]["data"]       = []

        forceField["angle"] = {}
        forceField["angle"]["type"]       = ["Bond3", "HarmonicAngular"]
        forceField["angle"]["labels"]     = ["id_i", "id_j", "id_k", "K", "theta0"]
        forceField["angle"]["data"]       = []


        index = 0
        lipidId = 0
        for i in range(int(nLipidsX)):
            for j in range(int(nLipidsY)):

                for leaflet in range(2):

                    for k,t in enumerate(lipid):
                        # The last tail particles is placed at z = sigmaT/2.0+zSpacing
                        x = i*dx + dx/2.0 - box[0]/2.0
                        y = j*dy + dy/2.0 - box[1]/2.0
                        if leaflet == 0: # Upper leaflet
                            z =   lipidZ[k]
                        else: # Lower leaflet
                            z = - lipidZ[k]

                        state["data"].append([index,[x,y,z]])
                        structure["data"].append([index,t,lipidId])

                        index += 1

                    H1 = index-1
                    H2 = index-2
                    T1 = index-3
                    T2 = index-4
                    T3 = index-5

                    # Bonds
                    forceField["bond"]["data"].append([H1,H2,kH1_H2_POPC,b0_H1_H2_POPC])
                    forceField["bond"]["data"].append([H2,T1,kH2_T1_POPC,b0_H2_T1_POPC])
                    forceField["bond"]["data"].append([T1,T2,kT1_T2_POPC,b0_T1_T2_POPC])
                    forceField["bond"]["data"].append([T2,T3,kT2_T3_POPC,b0_T2_T3_POPC])

                    # Angles
                    forceField["angle"]["data"].append([H1,H2,T1,kH1_H2_T1_POPC,theta0_H1_H2_T1_POPC])
                    forceField["angle"]["data"].append([H2,T1,T2,kH2_T1_T2_POPC,theta0_H2_T1_T2_POPC])
                    forceField["angle"]["data"].append([T1,T2,T3,kT1_T2_T3_POPC,theta0_T1_T2_T3_POPC])

                    lipidId += 1

        forceField["verletList"] = {}
        forceField["verletList"]["type"]       = ["VerletConditionalListSet", "intra_inter"]
        forceField["verletList"]["parameters"] = {"cutOffVerletFactor":1.2}

        forceField["WCA"] = {}
        forceField["WCA"]["type"]       = ["NonBonded", "WCAType1"]
        forceField["WCA"]["parameters"] = {"cutOffFactor": 2.5,
                                           "condition":"inter"}

        forceField["WCA"]["labels"] = ["name_i", "name_j", "epsilon", "sigma"]
        forceField["WCA"]["data"]   = []

        forceField["WCA"]["data"].append(["T","T",eps,sigmaT])
        forceField["WCA"]["data"].append(["T","H",eps,(sigmaT+sigmaH)/2.0])
        forceField["WCA"]["data"].append(["H","H",eps,sigmaH])

        # Chapuza for the moment
        forceField["LJ"] = {}
        forceField["LJ"]["type"]       = ["NonBonded", "LennardJonesType2"]
        forceField["LJ"]["parameters"] = {"cutOffFactor": 2.5,
                                          "condition":"inter"}

        forceField["LJ"]["labels"] = ["name_i", "name_j", "epsilon", "sigma"]
        forceField["LJ"]["data"]   = []

        forceField["LJ"]["data"].append(["T","T",eps,sigmaT])
        forceField["LJ"]["data"].append(["T","H",0.0,(sigmaT+sigmaH)/2.0])
        forceField["LJ"]["data"].append(["H","H",0.0,sigmaH])

        #Set model
        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)


    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        return sel



import sys, os

import logging

from . import simulationStepBase

class pairwiseForces(simulationStepBase):
    """
    Component name: pairwiseForces
    Component type: simulationStep

    Author: Pablo Palacios-Alonso
    Date: 11/03/2024

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath"},
                         requiredParameters  = {"outputFilePath"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        parameters = {}

        parameters["outputFilePath"] = params["outputFilePath"]
        parameters["mode"]           = "Pairwise_force"

        simulationStep = {
            name:{
                "type":["MechanicalMeasure","PairwiseForceMeasure"],
                "parameters":{**parameters}
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)






import sys, os

import logging

from . import simulationStepBase

class meanSquareDisplacement(simulationStepBase):
    """
    Component name: meanSquareDisplacement
    Component type: simulationStep

    Author: Pablo Diez-Silva
    Date: 04/06/2024

    This component measures the mean square displacement of the particles in the simulation.
    The mean square displacement is calculated as the average of the square of the distance between the particles and their initial position.

    :param outputFilePath: Path to the output file
    :type outputFilePath: str

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath"},
                         requiredParameters  = {"outputFilePath"},
                         availableSelections = {"selection"},
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        parameters = {}

        parameters["outputFilePath"] = params.get("outputFilePath")


        simulationStep = {
            name:{
              "type":["GeometricalMeasure","meanSquareDisplacement"],
              "parameters":{**parameters}
            }
        }

        ############################################################

        self.setGroup("selection")
        self.setSimulationStep(simulationStep)



